#!/usr/bin/env python

import click
from click import echo, secho, pass_context, command, argument, option
import dateutil
from tabulate import tabulate

from nextcodecli.utils import dumps


@command(help="List projects")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@pass_context
def projects(ctx, is_raw):
    projects = ctx.obj.service.get_projects()
    if is_raw:
        click.echo(dumps(projects))
        return

    fields = ['project_id', 'internal_name', 'create_date', 'csa_file_path']
    table = []
    for project in projects:
        lst = []
        for f in fields:
            v = project[f] or 'N/A'
            if f in ('create_date',):
                try:
                    v = dateutil.parser.parse(v).strftime("%Y-%m-%d %H:%M")
                except Exception:
                    v = ''
            lst.append(v)

        table.append(lst)
    tbl = tabulate(table, headers=fields)
    click.echo(tbl)



#!/usr/bin/env python

import click

from nextcodecli.utils import check_profile


@click.group()
@click.pass_context
def cli(ctx):
    """Root subcommand for pipelines functionality"""
    if ctx.obj.service is None:
        check_profile(ctx)
        ctx.obj.service = ctx.obj.client.service("pipelines")
        ctx.obj.session = ctx.obj.service.session


from nextcodecli.commands.pipelines.status import status
from nextcodecli.commands.pipelines.list import list_pipelines
from nextcodecli.commands.pipelines.jobs import jobs
from nextcodecli.commands.pipelines.job import job
from nextcodecli.commands.pipelines.run import run

cli.add_command(status)
cli.add_command(list_pipelines)
cli.add_command(jobs)
cli.add_command(job)
cli.add_command(run)



import sys, os
import click
from click import command, argument, pass_context, secho, echo
import logging
from tabulate import tabulate

import nextcode
from nextcode.utils import random_string
from nextcode.csa import CSASession
from nextcode.exceptions import AuthServerError, CSAError

from nextcodecli.utils import abort, dumps, print_tab
from nextcodecli.csa_import import process_sample, parse

log = logging.getLogger(__name__)


@click.group()
@click.option(
    '-u',
    '--username',
    envvar='CSA_USERNAME',
    prompt=True,
    help="User in the CSA instance, typically administrator",
)
@click.option(
    '-p',
    '--password',
    envvar='CSA_PASSWORD',
    prompt=True,
    help="Password for user",
    hide_input=True,
)
@pass_context
def cli(ctx, username, password):
    """
    Manage CSA users

    Requires the CSA admin username and passwords, which you can put into envivonment as CSA_USERNAME and CSA_PASSWORD
    """
    client = nextcode.Client()
    root_url = client.profile.root_url
    session = CSASession(root_url, username, password)
    log.info("Managing users on CSA server %s..." % session.csa_url)
    ctx.obj.session = session


@command(help="Create a user in the CSA instance")
@argument('user_name', nargs=1)
@argument('password', nargs=1)
@pass_context
def create_user(ctx, user_name, password):
    try:
        ctx.obj.session.create_user(user_name, password)
    except CSAError as ex:
        abort(ex)
    secho("User '%s' has been created" % (user_name), bold=True)


@command(help="Add a user to a project")
@argument('user_name', nargs=1)
@argument('project', nargs=1)
@click.option('-r', '--role', default="researcher", help='Role for this user')
@pass_context
def add_to_project(ctx, user_name, project, role):
    try:
        ctx.obj.session.add_user_to_project(user_name, project, role)
    except CSAError as ex:
        abort(ex)
    secho("User '%s' has been added to project %s" % (user_name, project), bold=True)


@command(help="Get a list of all projects in the CSA instance")
@click.option('-r', '--raw', 'is_raw', is_flag=True)
@pass_context
def projects(ctx, is_raw):
    projects = ctx.obj.session.get_project_names()
    if is_raw:
        echo(dumps(projects))
    else:
        table = tabulate([[p] for p in projects], headers=["Project Name"])
        echo(table)


@command(help="Get information about a CSA project")
@argument('project_name', nargs=1)
@click.option('-r', '--raw', 'is_raw', is_flag=True)
@pass_context
def project(ctx, project_name, is_raw):
    project = ctx.obj.session.get_project(project_name)
    if is_raw:
        echo(dumps(project))
        return

    print_tab("Name", project["name"])
    print_tab("Key", project["key"])
    print_tab("ID", project["id"])
    print_tab("Org", project["organization_key"])
    print_tab("State", project["state"])
    print_tab("Ref", project["reference_version"])
    try:
        credentials = ctx.obj.session.get_s3_credentials_for_project(project_name)
    except Exception as ex:
        abort(ex)
    buckets = []
    for bucket, creds in credentials.items():
        if not creds["project_name"]:
            bucket += " (system)"
        buckets.append(bucket)

    print_tab("S3 Buckets", ", ".join(buckets))


@click.command("import")
@click.argument("manifest_file", nargs=1)
@click.argument("project_name", nargs=1)
@click.option('-r', '--random', 'random_pn', is_flag=True, help="Randomize PN")
@pass_context
def import_sample(ctx, manifest_file, project_name, random_pn):
    if not os.path.isfile(manifest_file):
        abort(f"Could not open manifest file {manifest_file}")
    project = ctx.obj.session.get_project(project_name)
    if not project:
        abort(f"Project {project_name} not found")
    org_key = project['organization_key']

    manifest = parse(manifest_file)
    for sample_id, sample in manifest.items():
        subject_id = sample["subject_id"]
        if random_pn:
            sample_id += "-{}".format(random_string())
            subject_id = sample_id
        contents = {
            'sample': sample,
            'org_key': org_key,
            'sample_id': sample_id,
            'subject_id': subject_id,
            'project_name': project_name,
            'skip_processing_state': False,
        }
        process_sample(ctx.obj.session, contents)


@command(help="Create a new CSA project")
@argument('project_name', nargs=1)
@click.option(
    '-o', '--org', default="internal_org", help='Organization that the project belongs to'
)
@click.option('-r', '--ref', default="hg38", help='Reference data version')
@pass_context
def create_project(ctx, project_name, org, ref):
    existing_project = ctx.obj.session.get_project(project_name)
    if existing_project:
        abort(f"Project {project_name} already exists")
    try:
        ret = ctx.obj.session.create_project(project_name, org, ref)
    except CSAError as ex:
        abort(ex)
    secho(f"Project {project_name} has been created.")


cli.add_command(create_user)
cli.add_command(add_to_project)
cli.add_command(projects)
cli.add_command(project)
cli.add_command(import_sample)
cli.add_command(create_project)



#!/usr/bin/env python

import os
import sys

import click
from click import pass_context, command, option, secho, echo
import requests
import logging

from nextcode.config import get_profile_config, get_default_profile, create_profile
from nextcode.exceptions import InvalidProfile, InvalidToken
from nextcode.client import get_api_key
from nextcodecli.utils import abort, dumps

log = logging.getLogger(__name__)


@command()
@option('-u', '--username')
@option('-p', '--password')
@option('-r', '--realm', default='wuxinextcode.com')
@option(
    '-h',
    '--host',
    default=None,
    help="Host override if not using profile, e.g. platform.wuxinextcodedev.com",
)
@option(
    '-t',
    '--token',
    'is_token',
    is_flag=True,
    help="Return refresh token instead of writing into current profile",
)
@pass_context
def cli(ctx, username, password, realm, host, is_token):
    """
    Authenticate against keycloak.
    """
    try:
        config = get_profile_config()
    except InvalidProfile as ex:
        secho(str(ex), fg="red")
        abort("Please create a profile with: nextcode profile add [name]")
    profile_name = get_default_profile()

    if username and password:
        if not is_token:
            echo("Authenticating from commandline parameters")
        host = host or config["root_url"]
        try:
            api_key = get_api_key(host, username, password, realm=realm)
        except InvalidToken as ex:
            abort(ex)
        if is_token:
            click.echo(api_key)
            return
        create_profile(profile_name, api_key=api_key, root_url=host)
        click.secho("Profile {} has been updated with api key".format(profile_name), bold=True)

    else:
        if host:
            root_url = "https://%s" % host
        else:
            root_url = config["root_url"]
        login_server = root_url + "/api-key-service"

        if login_server:
            echo("Launching login webpage ==> Please authenticate and then press continue.")
            click.launch(login_server)
            click.pause()
        else:
            click.secho(
                "No login server configured. Please aquire a refresh_token from "
                "somewhere manually.",
                fg='yellow',
            )

        # Note: readline must be imported for click.prompt() to accept long strings. Don't ask me why.
        import readline

        api_key = click.prompt("API Key", type=str)
        try:
            create_profile(profile_name, api_key=api_key)
        except InvalidProfile as ex:
            abort(ex)



#!/usr/bin/env python

import click
from nextcodecli.utils import dumps, print_tab


@click.command(name='list', help="List available pipelines")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def list_pipelines(ctx, is_raw):
    session = ctx.obj.session
    resp = session.get(session.url_from_endpoint('pipelines'))
    resp.raise_for_status()
    pipelines = []
    for pipeline in resp.json()['pipelines']:
        resp = session.get(pipeline['links']['self'])
        pipelines.append(resp.json())
    if is_raw:
        click.echo(dumps(pipelines))
    else:
        for pipeline in pipelines:
            print_tab('name', pipeline['name'])
            print_tab('description', pipeline['description'])
            print_tab('author', pipeline['author'])
            print_tab('version', pipeline['version'])
            click.echo('')



#!/usr/bin/env python

import os, sys
import collections
import json
import uuid
import datetime
from operator import itemgetter
import subprocess
import tempfile
import click
import boto3
import dateutil

from nextcodecli.utils import (
    get_logger,
    dumps,
    abort,
    print_tab,
    print_error,
    print_warn,
    print_success,
)
from nextcodecli.pipelines.plot import write_instance_metrics_plot


class WXNCSecureShellError(Exception):
    pass


WORKER_NOT_READY_MSG = (
    "Could not connect to worker- the worker might not be up yet.\n" "Try again in a few minutes."
)
JOB_NOT_LAUNCHED_MSG = "Job has not instance."
BASTION_UNREACHABLE_MSG = (
    "Can not connect to worker - "
    "bastion host is not defined for profile.\n"
    "Check with your administrator whether SSH access "
    "to pipeline workers is enabled."
)


log = get_logger(__file__)


@click.group(help="Manage the worker EC2 instance (developer-only)")
@click.pass_context
def instance(ctx):
    instances = ctx.obj.job.get('instances')
    if not instances:  # None or [ ].
        abort(JOB_NOT_LAUNCHED_MSG)
    else:
        ctx.obj.instance_id = instances[0]['instance_id']


def get_instance_for_job(session, job_resource):
    link = session.links(job_resource)['instances']
    resp = session.get(link)
    resp.raise_for_status()
    try:
        instance_url = resp.json()[0]['links']['self']
    except IndexError:
        return None
    resp = session.get(instance_url)
    job_instance = resp.json()

    current_url = job_instance['links']['current']
    job_instance['current'] = {}
    try:
        resp = session.get(current_url)
        resp.raise_for_status()
        job_instance['current'] = resp.json()
    except Exception:
        pass

    # ! backwards-compatible hack since the details used to be a string from the API
    try:
        job_instance['details'] = json.loads(job_instance.get('details', '{}'))
    except Exception:
        pass
    return job_instance


@instance.command(
    'refresh',
    help="Force update the status of the job to reflect"
    "Instance tags. This is meant for local development",
)
@click.option('-s', '--status', help='Manually update the status of the job')
@click.pass_context
def instance_refresh(ctx, status):
    """This is only meant for local development or in cases where the sqs worker lambda
    is not correctly updating the status of the job in the pipelines-service.
    The duct-tools process on the worker updates the tag of the ec2 machine as well as
    notify the SQS service and we can poll for this flag to manually update the status
    in the pipelines-service.
    """
    session = ctx.obj.session
    job_resource = ctx.obj.job
    job_status = job_resource['status']
    events_url = job_resource['links']['events']
    if status:
        instance_status = status
    else:
        instance_resource = get_instance_for_job(session, job_resource)
        if not instance_resource or not instance_resource['current']:
            abort("Instance not found")

        current = instance_resource['current']
        instance_status = None
        for tag in current['Tags']:
            if tag['Key'] == 'job-status':
                instance_status = tag['Value']

    if instance_status != job_status:
        click.echo(
            "Instance has status '%s' and job has status '%s'" % (instance_status, job_status)
        )
        # create a 'fake' event to trigger the job to switch to the desired status
        # normally this event would come from the sqs worker lambda but in our localhost
        # case that lambda is not servicing the api
        event_guid = str(uuid.uuid4())
        event_name = ''
        if instance_status in ('RUNNING',):
            event_name = 'job_started'
        elif instance_status in ('FINISHED', 'FAILED', 'CANCELLED'):
            event_name = 'job_completed'
        body = {
            'event_name': event_name,
            'event_guid': event_guid,
            'event': {'status': instance_status},
            'date_time': datetime.datetime.utcnow().isoformat(),
        }
        resp = session.post(events_url, json=body)
        resp.raise_for_status()
        click.echo("Job status has been updated to '%s'" % instance_status)
    else:
        click.echo("Job status '%s' matches instance status" % job_status)


def test_connection(host_name, ssh_cmd):
    # Helper for testing ssh connectivity.

    log.debug("SSH test command for {host}: {cmd}".format(host=host_name, cmd=ssh_cmd))
    proc = subprocess.Popen(ssh_cmd.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stderr, stdout = proc.communicate()
    log.debug(
        "SSH test command returned: exit code: {code}stdout:{stdout}, stderr:{stdout}".format(
            code=proc.returncode, stdout=stdout, stderr=stderr
        )
    )
    if proc.returncode != 0:
        # sometimes errors come through stdout - who'da thunk it?
        reason = (stderr or stdout).decode().strip("\r\n")
        err_msg = (
            "Could not connect to {host_name}. Reason given:\n{reason} - "
            "Exit code: {exit_code}".format(
                host_name=host_name,
                exit_code=proc.returncode,
                reason=click.style(reason, bold=True),
            )
        )
        raise WXNCSecureShellError(err_msg)
    return proc.returncode


def initialize_ssh_keys(key_name=None):
    if key_name is None:
        log.debug("no keyname specified - trying default key")
        os.system("ssh-add")
        return
    try_paths = [os.getcwd(), "~/.ssh/"]
    for path in try_paths:
        key_file_path = "{path}/{key_name}".format(path=path, key_name=key_name)
        log.debug("Trying to add {path}".format(path=key_file_path))

        if os.path.isfile(key_file_path):
            # Ignore B605 - key_file_path is not a user-supplied variable.
            os.system("ssh-add {key_file_path}".format(key_file_path=key_file_path))
        else:
            log.debug("{path} not found or not a valid key".format(path=key_file_path))


# Display this  string if we have no idea what the error is.
SSH_ERROR_HELP = """{error_message}

Common causes for failures:

1) Feature is disabled.
   The pipeline worker ssh command is only meant for development environments.
   Please contact the administrator of your environment for further information.

2) SSH-key misconfiguration.
   The nextcode-cli requires two SSH keys to open a secure shell to the pipeline workers.
   These are:
    * The bastion SSH key
    * The pipeline worker ssh key

   Provided these keys, you should run the following commands before retrying.
    * Start the agent:
      $ eval `ssh-agent`
    * Add the keys
      $ ssh-add [path-to-bastion-SSH-key]
      $ ssh-add [path-to-pipeline-worker-SSH-key]
      """


@instance.command('ssh', help="SSH into a the running worker instance. Requires SSH key.")
@click.option(
    '-p',
    '--public',
    is_flag=True,
    help='Connect to the public IP of the instance instead of going through bastion',
)
@click.pass_context
def instance_ssh(ctx, public):
    session = ctx.obj.session
    svc = ctx.obj.service
    job_resource = ctx.obj.job
    job_id = job_resource['job_id']
    worker = get_instance_for_job(session, job_resource)
    if not worker:
        abort("Instance not found")

    if public:
        username = 'ubuntu'
        public_dns_name = worker.get('current', {}).get('PublicDnsName')
        if not public_dns_name:
            abort("Public DNS not found")
        key_name = worker.get('current', {}).get('KeyName')
        key_file = '~/.ssh/{}'.format(key_name)
        cmd = [
            'ssh',
            '-oStrictHostKeyChecking=no',
            '-i',
            key_file,
            '-t',
            '{}@{}'.format(username, public_dns_name),
        ]
        print(" ".join(cmd))
        subprocess.call(cmd)
        return

    try:
        bastion_host = svc.app_info.get('bastion_host')
        if bastion_host is None:
            raise abort(BASTION_UNREACHABLE_MSG)

        instance_id = worker['instance_id']
        if not worker['current']:
            abort(
                "Unable to connect. Instance {instance_id} has been terminated.".format(
                    instance_id=instance_id
                )
            )

        state = worker['current']['State']['Name']
        if state.lower() != 'running':
            err_str = "Instance '{instance_id}' is in state '{state}' which is not 'running'. \n"
            err_str += "Start with 'nextcode pipelines job {job_id} instance start'"
            abort(err_str.format(instance_id=instance_id, job_id=job_id, state=state.lower()))

        worker_ip_address = worker['current']['PrivateIpAddress']
        if not worker_ip_address:
            err_msg = WORKER_NOT_READY_MSG
            abort(err_msg.format(instance_id=instance_id))

        pipeline_path = '/opt/etl-pipelines/pipelines/%s' % job_resource['details']['pipeline_name']

        # Base templates for a worker and bastion ssh commands
        worker_ssh_cmd = " ssh -t {worker_options} {username}@{public_dns_name}{worker_cmd}"
        bastion_ssh_cmd = "ssh {bastion_options} -tA {bastion_host} {cmd}"
        key_name = worker.get('current', {}).get('KeyName')
        # Let's try to autodiscover the worker ssh key - it's a bit of a long shot, but it's free.
        initialize_ssh_keys(key_name)
        # While we're at it, let's attempt adding the default bastion key as well.
        initialize_ssh_keys()

        click.echo("Connecting to bastion host {host}... ".format(host=bastion_host), nl=False)

        # Create a version of the sshs string which immediately exit - for testing.
        # This allows us to give more fine-grained error messaging to the user.
        test_bastion_cmd = bastion_ssh_cmd.format(
            bastion_host=bastion_host,
            bastion_options="-oBatchMode=yes -oStrictHostKeyChecking=no",
            cmd='exit',
        )

        test_worker_ssh_cmd = worker_ssh_cmd.format(
            worker_options="-oBatchMode=yes -oStrictHostKeyChecking=no",
            username='ubuntu',
            public_dns_name=worker_ip_address,
            worker_cmd=' exit',
        )

        # Verify that we can connect, both to the bastion and the worker host.
        try:
            if test_connection('bastion host', test_bastion_cmd) == 0:
                print_success("SUCCESS")
        except WXNCSecureShellError as wex:
            if "Permission denied (publickey" in str(wex):
                raise WXNCSecureShellError(
                    "Unable to connect to bastion host due to wrong or missing SSH key"
                )
            else:
                raise
        try:
            test_connection(
                'worker instance', test_bastion_cmd.replace(' exit', test_worker_ssh_cmd)
            )
        except WXNCSecureShellError as wex:

            if "Permission denied (publickey" in str(wex):
                print_error(
                    "Unable to connect to worker due to wrong or missing SSH key",
                    "Key [{key_name}] needs to be added to your ssh-agent".format(
                        key_name=key_name
                    ),
                )
                print("Try this:")
                print("  eval `ssh-agent`")
                print("  ssh-add ~/.ssh/id_rsa")
                print("  ssh-add ~/.ssh/%s" % key_name)
                sys.exit(1)

            elif "ssh_exchange_identification: Connection closed by remote host" in str(wex):
                abort(WORKER_NOT_READY_MSG)
            else:
                raise

        worker_ssh_cmd = worker_ssh_cmd.format(
            public_dns_name=worker_ip_address,
            username='ubuntu',
            worker_options='-oStrictHostKeyChecking=no',
            worker_cmd=' /bin/bash',
        )

        # Construct the final ssh string.
        full_ssh_cmd = bastion_ssh_cmd.format(
            bastion_options='-oStrictHostKeyChecking=no',
            bastion_host=bastion_host,
            cmd=worker_ssh_cmd,
        )

        # print out the steps for reference
        try:
            click.echo("Job status:")
            for step in sorted(
                [s for s in job_resource['steps'] if s['step_identifier']],
                key=itemgetter('step_identifier'),
            ):
                print_tab(step['step_identifier'], step['step_name'])
            click.echo("# Pipeline path:\n  {pipeline_path}".format(pipeline_path=pipeline_path))
        except Exception as e:
            print_error(repr(e))

        click.echo("Connecting to worker ... ")
        exit_code = subprocess.call(full_ssh_cmd.split(" "))

        if exit_code not in [0, 127, 130]:  # Exit code 139 is Ctrl-C.
            # Exit code 127 sometimes pops up for no reason.
            raise WXNCSecureShellError(
                "Failed to connect to worker. Exit code {exit_code}".format(exit_code=exit_code)
            )
    except WXNCSecureShellError as shhex:
        abort(SSH_ERROR_HELP.format(error_message=shhex))


def start_or_stop_instance(ctx, action):
    session = ctx.obj.session
    job_resource = ctx.obj.job
    instance_resource = get_instance_for_job(session, job_resource)
    if instance_resource['details'].get('InstanceLifecycle') == 'spot':
        abort("Spot Instances cannot be started or stopped")
    if not instance_resource:
        abort("Instance not found")
    link = session.links(instance_resource)['self']
    resp = session.patch(link, json={'action': action})
    if resp.status_code != 200:
        abort("Cannot %s instance: %s" % (action, resp.json().get('message')))
    print(json.dumps(resp.json(), indent=4))
    resp.raise_for_status()


@instance.command('start', help="Start the worker instance for the job.")
@click.pass_context
def instance_start(ctx):
    start_or_stop_instance(ctx, 'start')


@instance.command('stop', help="Stop the worker instance for the job.")
@click.pass_context
def instance_stop(ctx):
    start_or_stop_instance(ctx, 'stop')


@instance.command('view', help="View instance information for the job.")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def instance_view(ctx, is_raw):
    """Quick and dirty instance viewer.
    """
    session = ctx.obj.session
    job_resource = ctx.obj.job
    instance_resource = get_instance_for_job(session, job_resource)
    if not instance_resource:
        abort("Instance not found")
    details = instance_resource['details']
    ami_details = instance_resource.get('ami_details', {})
    current = instance_resource['current']
    ami_version = None
    etl_pipelines_version = None
    for tag in ami_details.get('Tags', []):
        if tag['Key'] == 'version':
            ami_version = tag['Value']
        elif tag['Key'] == 'etl-pipelines-version':
            etl_pipelines_version = tag['Value']
    if is_raw:
        dct = {
            'instance_status_at_launch': details,
            'ami_details': ami_details,
            'current_instance_status': current,
        }
        click.echo(dumps(dct))
    else:
        rows = collections.OrderedDict()
        rows['Instance ID'] = details['InstanceId']
        rows['Instance Type'] = details['InstanceType']
        rows['SSH Key Name'] = details['KeyName']
        rows['Launch Time'] = details['LaunchTime']
        rows['AMI ID'] = details['ImageId']
        rows['AMI Version'] = ami_version
        rows['ETL Pipelines'] = etl_pipelines_version
        rows['Private IP Address'] = details['PrivateIpAddress']
        rows['Public IP Address'] = current.get('PublicIpAddress', 'N/A')
        rows['VPC ID'] = details['VpcId']
        if 'InstanceLifecycle' in details:
            rows['InstanceLifecycle'] = details['InstanceLifecycle']
        rows['Placement'] = current.get('Placement', {}).get('AvailabilityZone', 'Unknown')

        rows['Current Status'] = current.get('State', {}).get('Name', 'NOT FOUND')

        for k, v in rows.items():
            print_tab(k, v)


@instance.command('metrics', help="View metrics for instance")
@click.pass_context
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.option('-a', '--all', 'is_all', is_flag=True, help='Include all steps')
@click.option('-p', '--period', 'period', default=1, help='Minutes between data points')
@click.option('-w', '--width', 'image_width', default=1920, help='Width of the output image')
@click.option('-h', '--height', 'image_height', default=1024, help='Height of the output image')
def instance_metrics(ctx, is_raw, period, is_all, image_width, image_height):
    session = ctx.obj.session
    job_resource = ctx.obj.job
    job_id = job_resource['job_id']
    instance_resource = get_instance_for_job(session, job_resource)
    if not instance_resource:
        abort("Instance not found")
    metrics_url = instance_resource['links']['metrics']
    instance_id = instance_resource['instance_id']
    data = {'period': period}
    resp = session.get(metrics_url, data=data)
    if resp.status_code != 200:
        abort("Could not get metrics for instance %s: %s" % (instance_id, resp.text))
    max_num_points = 0
    metrics = []
    for metric_name, points in resp.json().items():
        metric = {
            "metric_name": metric_name,
            "output": tempfile.NamedTemporaryFile(mode='w', delete=False),
            "num_points": 0,
        }
        metrics.append(metric)
        results = {}
        for p in points:
            dt = dateutil.parser.parse(p['date_time'])
            results[dt] = p['value']

        metric['num_points'] = num_points = len(results)
        if num_points > 0:
            click.echo("%s - %s datapoints found" % (metric_name, num_points))
            if num_points > max_num_points:
                max_num_points = num_points

            left_trim = True
            for k in sorted(results.keys()):
                if left_trim and results[k] == 0:
                    continue
                left_trim = False
                line = "%s\t%s" % (k.strftime('%Y-%m-%d %H:%M'), results[k])
                if is_raw:
                    click.echo(line)
                else:
                    metric['output'].write(line + '\n')
            metric['output'].close()

    if max_num_points == 0:
        print_warn("No metrics found for instance %s" % instance_id)
        return
    if is_raw:
        return
    file_name = write_instance_metrics_plot(
        job_id,
        instance_id,
        metrics,
        job_resource['steps'],
        image_width,
        image_height,
        max_num_points,
        is_all,
    )
    if file_name:
        click.secho("Graph will be produced here: %s" % file_name, fg='green', bold=True)



#!/usr/bin/env python

import click
from nextcode.config import get_default_profile
import nextcodecli
from nextcodecli.utils import dumps, print_table
from ...utils import abort, fmt_date


@click.command(help="Show the status of the project api")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def status(ctx, is_raw):
    svc = ctx.obj.service
    profile_name = get_default_profile()
    if is_raw:
        dct = svc.session.root_info
        dct.update({'version': nextcodecli.get_version_string(), 'profile': profile_name})
        click.echo(dumps(dct))
    else:
        click.echo(nextcodecli.get_version_string() + '\n')

        click.echo("{0:20}{1}".format('profile', profile_name))
        click.echo('')
        click.echo("Contacting project api '%s'... " % svc.base_url, nl=False)

        click.echo("Success!\n\nAPI deployment information:")
        print_table(svc.build_info)

        svc = ctx.obj.service
        click.echo("{0:20}{1}".format("Current project", svc.project_name))
        click.echo("{0:20}{1}".format("Admin", "Yes" if svc.current_user["admin"] else "No"))
        click.echo("{0:20}{1}".format("Minio Url", svc.session.root_info["app_info"]["minio_url"]))
        click.echo('')


@click.command(help="Create the current project in the project api")
@click.pass_context
def create(ctx):
    svc = ctx.obj.service
    try:
        project = svc.create_project()
    except Exception as e:
        abort(e)
    click.echo(dumps(project))


@click.command(help="Update the current project in the project api")
@click.option('-t', '--title', help='Project title')
@click.option('-d', '--description', help='Project description')
@click.pass_context
def update(ctx, title, description):
    svc = ctx.obj.service
    project = svc.project
    data = {}
    if title:
        data["title"] = title
    if description:
        data["description"] = description
    try:
        project = svc.session.put(project["links"]["self"], json=data)
    except Exception as e:
        abort(e)
    click.echo(dumps(project.json()))


@click.command(help="View project information")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def view(ctx, is_raw):
    svc = ctx.obj.service
    project = svc.project
    if is_raw:
        click.echo(dumps(project))
        return
    if not project:
        abort(f"Project {svc.project_name} not found. Please run 'nextcode project create'")
    click.echo("{0:20}{1}".format("Project name", click.style(project["project_name"], bold=True)))
    click.echo("{0:20}{1}".format("Project ID", project["project_id"]))
    click.echo("{0:20}{1}".format("Title", project["title"]))
    click.echo("{0:20}{1}".format("Description", project["description"]))
    click.echo("{0:20}{1}".format("Create date", fmt_date(project["create_date"])))



import click
import os, sys
import logging

from .utils import console_handler, abort

help_string = """A utility for interfacing with WuXi Nextcode services.

This tool allows you to communicate with the pipelines service, CSA, workflow service and GOR Query API.
For all usage you will need to authenticate against the specific service profile you are using.

Please look at the subcommands below for details.
"""
plugin_folder = os.path.join(os.path.dirname(__file__), 'commands')
profile_check = False


class CLI(click.MultiCommand):
    def __init__(self, *args, **kwargs):
        click.MultiCommand.__init__(self, *args, **kwargs)
        opt = click.Option(
            param_decls=["--verbose", "-v"],
            help='Output logs for debugging',
            type=click.Choice(['debug', 'warning', 'error', 'info']),
        )
        self.params.append(opt)
        opt = click.Option(
            param_decls=["--profile", "-p"], help='Use a specific profile for this command'
        )
        self.params.append(opt)

    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(plugin_folder):
            if filename.startswith("__"):
                continue
            if filename.endswith('.py'):
                rv.append(filename[:-3])
            elif os.path.isdir(os.path.join(plugin_folder, filename)):
                rv.append(filename)
        rv.sort()

        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(plugin_folder, name)
        # if the command is a folder we look for the entrypoint in the __init__ file
        if os.path.isdir(fn):
            fn = os.path.join(fn, '__init__.py')
        else:
            fn = os.path.join(plugin_folder, name + '.py')

        if not os.path.exists(fn):
            abort("Command {} not found".format(name))

        with open(fn) as f:
            code = compile(f.read(), fn, 'exec')
            # We are only passing our own code to eval
            # No unchecked external code should ever be passed to eval
            eval(code, ns, ns)  # nosec
        cmd = ns.get('cli', None)
        return cmd


class Context(object):
    def __init__(self):
        self.verbose = False
        self.profile_name = None
        self.home = os.getcwd()
        self.session = None
        self.service = None
        self.client = None


@click.command(cls=CLI, help=help_string)
@click.pass_context
def cli(ctx, verbose=None, profile=None):
    """Convert files using specified converter"""
    # by default we log out to console WARN and higher but can view info with -v
    # make sure this is called before importing the sdk so that we can grab all logging
    if verbose:
        console_handler.setLevel(getattr(logging, verbose.upper()))

    from nextcode.client import Client
    from nextcode.exceptions import InvalidProfile

    if ctx.obj is None:
        ctx.obj = Context()
        ctx.obj.verbose = verbose
        ctx.obj.profile_name = profile
        try:
            ctx.obj.client = Client(profile=profile)
        except InvalidProfile:
            pass


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python

"""
Contains gnuplot methods for drawing graphs. Assumes macos and is generally a bit hacky.
At least we keep it all in one place.
"""
import os
import sys
import subprocess
import datetime
import dateutil.parser

import click

from ..utils import print_error


def write_instance_metrics_plot(
    job_id, instance_id, metrics, job_steps, width, height, max_num_points, is_all=False
):

    # check if gnuplot is installed
    try:
        p = subprocess.Popen(
            ['gnuplot', '-e', 'set terminal cairo'], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = p.communicate()
        if p.returncode != 0:
            raise Exception(err)
    except Exception as e:
        if sys.platform != "darwin":
            print_error("This method only supports macOS. Please use --raw option.")
            sys.exit(1)
        print_error(str(e))
        print_error("Gnuplot needs to be installed with cairo extension to view metrics.")
        click.echo("Please install with 'brew reinstall gnuplot --with-cairo'")
        sys.exit(1)
    proc = subprocess.Popen(
        ['gnuplot', '-p'], shell=True, stdin=subprocess.PIPE, close_fds=True  # Ignore B602
    )
    file_name = os.path.expanduser('~/job_%s.png' % job_id)
    try:
        os.remove(file_name)
    except Exception:
        pass

    input_string = """set datafile separator "\t"
set title "job {job_id} ({instance_id})"
set term pngcairo size {width}, {height}
set ylabel "Percent"
set xlabel "Time"
set xdata time
set timefmt "%Y-%m-%d %H:%M"
set format x "%H:%M"
set key left top
set grid
set output '{file_name}'
set yrange [0:100]
        """.format(
        job_id=job_id, instance_id=instance_id, file_name=file_name, width=width, height=height
    )
    proc.stdin.write(input_string.encode('UTF-8'))

    steps = []
    for step in job_steps:
        date_started = step.get('date_started')
        if not date_started or ('load' in step['step_name'] and not is_all):
            continue
        date_started = dateutil.parser.parse(date_started)
        date_started = date_started.replace(tzinfo=None, second=0, microsecond=0)
        steps.append((date_started, True, step['step_name'] + ' start'))
        if step.get('date_completed'):
            date_completed = dateutil.parser.parse(step.get('date_completed'))
            date_completed = date_completed.replace(tzinfo=None, second=0, microsecond=0)
            steps.append((date_completed, False, step['step_name'] + ' end'))

    arrow_commands = ""
    min_minutes = max(1, min(max_num_points / 40, 5))
    min_date = None
    for dt, is_start, name in sorted(steps):
        if min_date and dt < min_date + datetime.timedelta(minutes=min_minutes):
            dt = min_date + datetime.timedelta(minutes=min_minutes)
        min_date = dt
        col = [(3, '#000088'), (2, '#008800')][is_start]
        dt_label = dt + datetime.timedelta(seconds=90)
        arrow_commands += """set arrow from '{dt}', graph 1.0 to '{dt}', graph 0.7 ls {col1}
set label "{step}" at '{dt_label}',99 rotate right font "Arial,10" noenhanced front tc rgb "{col2}"
""".format(
            dt=dt, dt_label=dt_label, step=name, col1=col[0], col2=col[1]
        )
    arrow_commands = (arrow_commands + '\n').encode('UTF-8')
    proc.stdin.write(arrow_commands)

    plot_commands = "plot "
    for metric in metrics:
        if metric['num_points'] == 0:
            continue
        cmd = """"{filename}" using 1:2 with lines lw 2 title '{metric}', """.format(
            filename=metric['output'].name, metric=metric['metric_name']
        )
        plot_commands += cmd

    plot_commands = plot_commands[:-2]
    plot_commands = (plot_commands + '\n').encode('UTF-8')
    proc.stdin.write(plot_commands)
    return file_name



import click


@click.command("setproject")
@click.argument('project', nargs=1)
@click.pass_context
def cli(ctx, project):
    """
    Set the current project.

    This command will set the project context in the client and save it
    into the current profile. A project is required to run any gor query
    or to inspect already run queries.
    """
    ctx.obj.client.profile.project = project
    click.secho("Project has been set to: {}".format(project), bold=True)



#!/usr/bin/env python

import click
from nextcode.config import get_default_profile
import nextcodecli
from nextcodecli.utils import dumps, print_table


@click.command(help="Show the status of the gor query api")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def status(ctx, is_raw):
    svc = ctx.obj.service
    profile_name = get_default_profile()
    if is_raw:
        dct = {'version': nextcodecli.get_version_string(), 'profile': profile_name}
        for key in ('build_info', 'app_info', 'endpoints', 'current_user'):
            dct[key] = getattr(svc.session, key, None)
        click.echo(dumps(dct))
    else:
        click.echo(nextcodecli.get_version_string() + '\n')

        click.echo("{0:20}{1}".format('profile', profile_name))
        click.echo('')
        click.echo("Contacting gor query api '%s'... " % svc.base_url, nl=False)

        click.echo("Success!\n\nAPI deployment information:")
        print_table(svc.build_info)

        svc = ctx.obj.service
        click.echo("{0:20}{1}".format("Current project", svc.project))
        click.echo('')



#!/usr/bin/env python

import json
import click
from click import echo, secho, pass_context, command, argument, option

from nextcode.exceptions import ServerError

from nextcodecli.utils import dumps, print_table, print_tab
from nextcodecli import get_version_string


@command(help="Show the status of the workflow service")
@option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@pass_context
def status(ctx, is_raw):
    svc = ctx.obj.service
    session = ctx.obj.service.session
    profile_config = svc.client.profile
    profile_name = svc.client.profile_name
    if is_raw:
        dct = {
            'version': get_version_string(),
            'profile': profile_name,
            'profile_config': profile_config.content,
        }
        for key in ('build_info', 'app_info', 'endpoints', 'current_user'):
            dct[key] = getattr(session, key, None)
        resp = session.get(session.url_from_endpoint('pipelines'))
        resp.raise_for_status()
        dct['pipelines'] = [p['name'] for p in resp.json()['pipelines']]
        click.echo(dumps(dct))
    else:
        click.echo(get_version_string() + '\n')

        click.echo("Contacting workflow service '%s'... " % svc.base_url, nl=False)

        click.echo("Success!\n\nAPI deployment information:")
        print_table(svc.build_info)
        print_table(svc.app_info)

        click.echo('')


@command(help="Run a smoketest of the workflow service")
@option(
    '--test',
    type=click.Choice(['viewscm', 'listwork', 'nextflowinfo', 'nextflowrun']),
    default='nextflowrun',
    show_default=True,
    help="Name of the test to run",
)
@option(
    '--timeout',
    type=int,
    default=30,
    show_default=True,
    help="Number of seconds to wait for test to complete",
)
@pass_context
def smoketest(ctx, test, timeout):
    session = ctx.obj.session
    url = session.url_from_endpoint('health') + 'smoketest'
    secho("Running smoketest by calling POST %s... This might take a while." % url, bold=True)
    try:
        resp = session.post(url, json={'test': test, 'timeout': timeout})
    except ServerError as e:
        secho("Smoketest failed!", fg='red', bold=True)
        r = e.response
        if not r:
            echo(repr(e))
            return
        try:
            error = r.get('error', {})
            pod_name = error.get('pod_name')
            state = json.loads(error.get('state', '{}'))
            exit_code = state.get('_exit_code')
            print_tab("Error code", r.get('code'))
            print_tab("Error description", error.get('description'))
            if pod_name:
                print_tab("Pod name", pod_name)
            if exit_code:
                print_tab("Exit code", exit_code)
            if state:
                print_tab("Reason", state.get('_reason'))
                state_txt = ""
                try:
                    state_txt = ['%s: %s' % (k, v) for k, v in state.items() if v][0]
                except:
                    pass
                print_tab("State", state_txt)
                print_tab("Container message", state.get('_message'))

            if pod_name and not exit_code:
                echo(
                    "You can investigate this issue further by running 'kubectl describe pod %s -n gorkube'"
                    % pod_name
                )
        except Exception as e:
            raise
            echo(dumps(r))
    else:
        secho("Smoketest succeeded!", bold=True, fg='green')
        echo(
            "Nextflow pod completed successfully and produced the following output:\n%s"
            % "\n".join(resp.json()['logs'])
        )



#!/usr/bin/env python

import click

from functools import wraps

from nextcode import get_service
from nextcode.exceptions import ServerError, InvalidProfile
from nextcodecli.utils import print_error, abort, check_profile


def check_project(func):
    @wraps(func)
    def _check(*args, **kwargs):
        ctx = click.get_current_context()
        svc = ctx.obj.service
        if not svc.project:
            abort("Please set a project by running: nextcode setproject [name]")
        else:
            return func(*args, **kwargs)

    return _check


@click.group()
@click.pass_context
def cli(ctx):
    """Root subcommand for project api functionality.
    """
    if ctx.obj.service is None:
        check_profile(ctx)
        ctx.obj.service = ctx.obj.client.service("project")


from nextcodecli.commands.project.fileaccess import (
    list_bucket,
    upload,
    download,
    delete,
    head,
    credentials,
)
from nextcodecli.commands.project.status import status, create, update, view
from nextcodecli.commands.project.users import users

cli.add_command(list_bucket)
cli.add_command(upload)
cli.add_command(download)
cli.add_command(delete)
cli.add_command(head)
cli.add_command(status)
cli.add_command(users)
cli.add_command(create)
cli.add_command(update)
cli.add_command(view)
cli.add_command(credentials)



from tabulate import tabulate
import dateutil
import datetime
import time
import click


process_status_to_color = {
    'SUBMITTED': 'yellow',
    'STARTED': 'yellow',
    'ABORTED': 'white',
    'FAILED': 'red',
    'COMPLETED': 'green',
}


def get_limits_string(process):
    try:
        limits = process['details']['pod']['resources']['limits']
    except (IndexError, KeyError, TypeError):
        limits = {}
    if limits:
        return ", ".join(["%s: %s" % (k, v) for k, v in limits.items()])
    else:
        return ""


def get_processes_table(processes, is_color=True, is_wide=False):
    headers = ['id', 'name', 'container', 'status', 'submit_date', 'duration']
    if is_wide:
        headers.extend(['start_date', 'complete_date', 'runtime', 'limits', 'pod', 'node'])
    table = []
    for process in processes:
        submit_date = 'N/A'
        duration = process['duration']
        runtime = 'N/A'
        if process['submit_date']:
            dt = dateutil.parser.parse(process['submit_date'])
            submit_date = dt.strftime('%Y-%m-%d %H:%M:%S')
            if not duration and process['status'] == 'STARTED':
                duration = (datetime.datetime.utcnow() - dt).total_seconds() * 1000
        start_date = None
        start_date_string = 'N/A'
        complete_date = None
        complete_date_string = 'N/A'
        exit_code = ''
        if process['start_date']:
            start_date = dateutil.parser.parse(process['start_date'])
            start_date_string = start_date.strftime('%Y-%m-%d %H:%M:%S')

        if process['complete_date'] and start_date:
            complete_date = dateutil.parser.parse(process['complete_date'])
            complete_date_string = complete_date.strftime('%Y-%m-%d %H:%M:%S')
            exit_code = process['exit_code']
            runtime = complete_date - start_date
            runtime = str(runtime).split('.')[0]
        elif start_date:
            runtime = (datetime.datetime.utcnow() - start_date).total_seconds() * 1000
            runtime = datetime.timedelta(seconds=(runtime // 1000))
            runtime = str(runtime).split('.')[0]

        if exit_code:
            exit_code = click.style(' (exit code %s)' % exit_code, fg='red')
        else:
            exit_code = ''
        if duration:
            duration = datetime.timedelta(seconds=(duration // 1000))
            duration = str(duration).split('.')[0]
        node_name = 'N/A'
        try:
            node = process['details']['node']
            node_name = node['name']
            instance_type = node['labels']['beta.kubernetes.io/instance-type']
            node_name += " (%s)" % instance_type
        except (TypeError, KeyError):
            pass
        status = process['status'] + exit_code
        if is_color:
            status = click.style(status, fg=process_status_to_color.get(process['status'], 'white'))
        row = [
            process['process_id'],
            process['name'][:50],
            process['container'],
            status,
            submit_date,
            duration,
        ]
        if is_wide:
            row.extend(
                [
                    start_date_string,
                    complete_date_string,
                    runtime,
                    get_limits_string(process),
                    process['native_id'],
                    node_name,
                ]
            )
        table.append(row)

    tbl = tabulate(table, headers=headers)
    return tbl



import datetime
import click
from click import command, argument, pass_context, secho, echo
import logging
from tabulate import tabulate

import nextcode
from nextcode.keycloak import KeycloakSession
from nextcode.csa import CSASession
from nextcode.exceptions import AuthServerError, InvalidProfile
from nextcodecli.utils import abort, dumps, print_tab

log = logging.getLogger(__name__)

CLIENT_ID = 'api-key-client'


def _get_user(ctx, user_name):
    try:
        user = ctx.obj.session.get_user(user_name)
    except AuthServerError as ex:
        abort(ex)
    return user


@click.group()
@click.option(
    '-u',
    '--username',
    default='admin',
    help="Administrator user in the keycloak instance",
    show_default=True,
)
@click.option(
    '-p',
    '--password',
    envvar='KEYCLOAK_PASSWORD',
    prompt=True,
    help="Password for administrator user",
    hide_input=True,
)
@click.option(
    '-r', '--realm', default='wuxinextcode.com', help="Keycloak realm to manage", show_default=True
)
@click.option(
    '-s', '--server', 'root_url', envvar='KEYCLOAK_SERVER', help="Keycloak server to manage"
)
@pass_context
def cli(ctx, username, password, realm, root_url):
    """
    Manage keycloak users

    Requires the keycloak admin password, which you can put into envivonment as KEYCLOAK_PASSWORD.

    By default the command uses the keycloak server from the current profile's API Key but can be overridden
    with KEYCLOAK_SERVER environment variable.
    """
    if not root_url:
        try:
            client = nextcode.Client()
        except InvalidProfile as ex:
            secho(str(ex), fg="red")
            abort("You can use --server to pick a keycloak server.")
        root_url = client.profile.root_url
    try:
        session = KeycloakSession(root_url, username, password, realm)
    except AuthServerError as ex:
        abort(ex)
    log.info("Managing users on keycloak server %s..." % session.auth_server)
    ctx.obj.session = session


@command(help="List all keycloak users in the realm")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@pass_context
def users(ctx, is_raw):
    users = ctx.obj.session.get_users()
    if is_raw:
        echo(dumps(users))
        return
    fields = ['username', 'email', 'name', 'enabled', 'created']
    table = []
    for user in users:
        table.append(
            [
                user['username'],
                user['email'],
                user.get('firstName', '') + ' ' + user.get('lastName', ''),
                user['enabled'],
                datetime.datetime.fromtimestamp(user['createdTimestamp'] // 1000),
            ]
        )
    tbl = tabulate(sorted(table), headers=fields)
    echo(tbl)


@command(help="View information about a keycloak user")
@argument('user_name', nargs=1)
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@pass_context
def user(ctx, user_name, is_raw):
    user = _get_user(ctx, user_name)
    roles = ctx.obj.session.get_user_roles(user_name)

    if is_raw:
        echo(dumps(user))
        echo(dumps(roles))
        return
    print_tab('Username', user['username'])
    print_tab('Email', user['email'])
    print_tab('Enabled', user['enabled'])
    print_tab('Name', "%s %s" % (user.get('firstName') or '', user.get('lastName') or ''))
    print_tab('Created', datetime.datetime.fromtimestamp(user['createdTimestamp'] // 1000))
    if user.get('federatedIdentities'):
        print_tab('Federation', user['federatedIdentities'][0]['identityProvider'])
    else:
        print_tab('Federation', '(Local user)')
    available_roles = ctx.obj.session.get_available_roles_for_user(user_name)
    if available_roles:
        echo("\nThe following roles can be added to the user: %s" % ', '.join(available_roles))


@command(help="Add a new keycloak user")
@argument('user_name', nargs=1)
@click.option('-p', '--new_password', help='Password for new user', prompt=True)
@pass_context
def add_user(ctx, user_name, new_password):
    try:
        ctx.obj.session.create_user(user_name, new_password)
    except AuthServerError as ex:
        abort(ex)
    echo("Success! User has been created. Please use the following authentication information:\n")
    echo(
        "Username: %s\nPassword: %s"
        % (click.style(user_name, bold=True), click.style(new_password, bold=True))
    )
    echo("")


@command(help="Reset the password of a keycloak user")
@argument('user_name', nargs=1)
@click.option('-p', '--new_password', help='Password for new user', prompt=True)
@pass_context
def reset_password(ctx, user_name, new_password):
    _ = _get_user(ctx, user_name)
    ctx.obj.session.set_user_password(user_name, new_password)
    echo("Password has been reset. Please use the following authentication information:\n")
    echo(
        "Username: %s\nPassword: %s"
        % (click.style(user_name, bold=True), click.style(new_password, bold=True))
    )
    echo("")


@command(help="Remove a keycloak user")
@argument('user_name', nargs=1)
@pass_context
def remove_user(ctx, user_name):
    _ = _get_user(ctx, user_name)
    session = ctx.obj.session
    session.delete_user(user_name)
    secho("User '%s' has been deleted" % (user_name), bold=True)


@command(help="Add a new role to a keycloak user")
@argument('user_name', nargs=1)
@argument('role_name', nargs=1)
@pass_context
def add_role(ctx, user_name, role_name):
    _ = _get_user(ctx, user_name)
    session = ctx.obj.session
    role_name = role_name.lower()
    user_roles = session.get_user_roles(user_name)
    if role_name in user_roles:
        abort("User '%s' already has role '%s'" % (user_name, role_name))

    available_roles = session.get_available_roles_for_user(user_name)
    if role_name not in available_roles:
        abort("Role '%s' is not available for user '%s'" % (role_name, user_name))

    ctx.obj.session.add_role_to_user(user_name, role_name)
    secho("Role '%s' has been added to user '%s'" % (role_name, user_name), bold=True, fg='green')


@command(help="Remove a role from a keycloak user")
@argument('user_name', nargs=1)
@argument('role_name', nargs=1)
@pass_context
def remove_role(ctx, user_name, role_name):
    _ = _get_user(ctx, user_name)
    session = ctx.obj.session
    role_name = role_name.lower()
    try:
        session.remove_role_from_user(user_name, role_name)
    except AuthServerError as ex:
        abort(ex)
    secho(
        "Role '%s' has been removed from user '%s'" % (role_name, user_name), bold=True, fg='green'
    )


@command(help="Create a new role on the server")
@argument('role_name', nargs=1)
@pass_context
def create_server_role(ctx, role_name):
    session = ctx.obj.session
    role_name = role_name.lower()
    try:
        session.create_role(role_name)
    except AuthServerError as ex:
        abort(ex)
    secho("Role '%s' has been created on the server" % (role_name), bold=True, fg='green')


@command(help="Create a new role on the server")
@argument('role_name', nargs=1)
@pass_context
def delete_server_role(ctx, role_name):
    session = ctx.obj.session
    role_name = role_name.lower()
    try:
        session.delete_role(role_name)
    except AuthServerError as ex:
        abort(ex)
    secho("Role '%s' has been deleted from the server" % (role_name), bold=True, fg='green')


@command(help="Add a new CSA user in a project")
@argument('user_name', nargs=1)
@argument('password', nargs=1)
@argument('project', nargs=1)
@click.option(
    '-u',
    '--admin-username',
    envvar='CSA_USERNAME',
    prompt=True,
    help="Username for CSA administrator user",
    hide_input=True,
)
@click.option(
    '-p',
    '--admin-password',
    envvar='CSA_PASSWORD',
    prompt=True,
    help="Password for CSA administrator user",
    hide_input=True,
)
@pass_context
def add_csa_user(ctx, user_name, password, project, admin_username, admin_password):
    client = nextcode.Client()
    root_url = client.profile.root_url
    try:
        session = CSASession(root_url, admin_username, admin_password)
    except AuthServerError as ex:
        abort(ex)

    try:
        session.create_user(user_name, password)
    except AuthServerError as ex:
        click.echo(str(ex))

    session.add_user_to_project(user_name, project, exist_ok=True)
    click.secho("User {} is now available in CSA project {}".format(user_name, project), bold=True)


cli.add_command(users)
cli.add_command(user)
cli.add_command(add_user)
cli.add_command(remove_user)
cli.add_command(reset_password)
cli.add_command(add_role)
cli.add_command(remove_role)
cli.add_command(create_server_role)
cli.add_command(delete_server_role)
cli.add_command(add_csa_user)



#!/usr/bin/env python
import click
import os
import json
from tabulate import tabulate

from nextcode.exceptions import ServerError
from nextcode.services.project.service import fmt_size
from ...utils import get_logger, abort, print_tab, dumps
from . import check_project

log = get_logger(name='commands.project', level='INFO')


@click.command("credentials")
@click.option('--create', is_flag=True, default=False, help='Create new credentials')
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
def credentials(ctx, create, is_raw):
    svc = ctx.obj.service
    try:
        ret = svc.get_credentials(create)
    except Exception as ex:
        abort(str(ex))
    if is_raw:
        print(json.dumps(ret, default=str))
        return
    else:
        click.echo("{0:22}{1}".format('AWS_ACCESS_KEY_ID', ret["aws_access_key_id"]))
        click.echo("{0:22}{1}".format('AWS_SECRET_ACCESS_KEY', ret["aws_secret_access_key"]))


@click.command("list")
@click.argument('path', required=False)
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
@check_project
def list_bucket(ctx, path, is_raw):
    """
    List files in the project tree
    """
    if not path:
        path = ""
    svc = ctx.obj.service
    ret = svc.list(path, raw=True)

    if is_raw:
        print(json.dumps(ret, default=str))
        return

    lst = []
    for r in ret:
        r["size"] = fmt_size(r["size"])
        try:
            r["modified"] = r["modified"].strftime("%b %d %H:%M")
        except Exception:
            pass
        if r["type"] == "file":
            r["name"] = r["name"].split("/")[-1]
        else:
            r["name"] = r["name"].split("/")[-2] + "/"
        lst.append([r["type"], r["size"], r["modified"], r["name"]])

    txt = tabulate(lst, headers=["type", "size", "modified", "name"])
    click.secho(f"Contents of {path}:", bold=True)
    print(txt)


@click.command()
@click.argument('filename', required=True)
@click.argument('path', required=True)
@click.pass_context
@check_project
def upload(ctx, filename, path):
    """
    Upload a file into the project tree
    """
    svc = ctx.obj.service
    try:
        remote_path = svc.upload(filename, path)
        click.secho(f"file has been uploaded to {remote_path}", bold=True)
    except Exception as e:
        log.debug(f"Error uploading {filename} to {path}: {e}")
        abort(str(e))


@click.command()
@click.argument('filename', required=True)
@click.argument('path', required=False)
@click.pass_context
@check_project
def download(ctx, filename, path):
    """
    Download a file from the project tree
    """
    if not path:
        path = "."
    svc = ctx.obj.service
    try:
        local_path = svc.download(filename, path)
        click.secho(f"file has been downloaded to {local_path}", bold=True)
    except Exception as e:
        log.debug(f"Error downloading {filename}: {e}")
        abort(str(e))


@click.command()
@click.argument('filename', required=True)
@click.option('--bytes', '-n', 'num_bytes', default=1024, help='Number of bytes to return')
@click.pass_context
@check_project
def head(ctx, filename, num_bytes):
    """
    Show the first n bytes from a file
    """
    svc = ctx.obj.service
    # TODO: Move into sdk
    bucket = svc.get_project_bucket()
    client = bucket.meta.client
    start_byte = 0
    stop_byte = num_bytes
    resp = client.get_object(
        Bucket=bucket.name, Key=filename, Range='bytes={}-{}'.format(start_byte, stop_byte)
    )
    contents = resp['Body'].read()
    try:
        # skip the last line since it's likely incomplete
        lines = contents.decode().split("\n")
        print("\n".join(lines[:-1]))
    except Exception:
        print(contents)


@click.command()
@click.argument('filename', required=True)
@click.pass_context
@check_project
def delete(ctx, filename):
    """
    Delete a file in the project tree
    """
    raise NotImplementedError("It is not a good idea to delete things.")
    svc = ctx.obj.service
    # TODO: Move into sdk?
    bucket = svc.get_project_bucket()
    resp = bucket.delete_objects(Delete={"Objects": [{"Key": filename}]})



#!/usr/bin/env python

import click
from click import command, option, pass_context
from tabulate import tabulate

from nextcode.exceptions import ServerError

from nextcodecli.commands.workflow import status_to_color
from nextcodecli.utils import get_logger, abort, dumps

log = get_logger()


@command(help="List jobs")
@option('--mine', 'is_mine', is_flag=True, help='List jobs from me')
@option('-u', '--user', 'user_name', help='User to filter for')
@option('-s', '--status', 'status', default=None, help='Filter status')
@option('-p', '--project', 'project', default=None, help='Filter by project name')
@option('--pipeline', 'pipeline', default=None, help='Filter by pipeline name')
@option('--page', 'is_page', is_flag=True, help='Page results')
@option('--running', 'is_running', is_flag=True, help='Only show jobs in pending/running state')
@option('--context', help='Filter by context')
@option(
    '--completed',
    'is_completed',
    is_flag=True,
    help='Only show jobs in failed/cancelled/completed state',
)
@option('-n', '--num', default=20, help='Maximum number of jobs to return')
@option(
    '-o',
    '--output',
    type=click.Choice(['normal', 'wide', 'json']),
    default='normal',
    help='Format output',
)
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@pass_context
def jobs(
    ctx,
    is_mine,
    user_name,
    status,
    project,
    pipeline,
    is_page,
    is_running,
    context,
    is_completed,
    num,
    output,
    is_raw,
):
    svc = ctx.obj.service
    if is_mine:
        try:
            user_name = svc.current_user['email']
        except KeyError:
            abort("You appear not to be logged in")

    state = None
    if is_running:
        state = "running"
    elif is_completed:
        state = "completed"
    try:
        jobs = svc.get_jobs(
            user_name,
            status,
            project=project,
            pipeline=pipeline,
            limit=num,
            state=state,
            context=context,
        )
    except ServerError as e:
        abort(str(e))
    is_wide = output == 'wide'
    if is_raw or (output == 'json'):
        click.echo(dumps(jobs))
        return

    fields = [
        'job_id',
        'pipeline_name',
        'user_name',
        'project_name',
        'submit_date',
        'duration',
        'status',
    ]
    if is_wide:
        fields.extend(['desc', 'context', 'complete_date', 'pod', '#proc'])
    jobs_list = []
    for job in jobs:
        job_list = []
        for f in fields:
            v = None
            if hasattr(job, f):
                v = getattr(job, f) or 'N/A'
            if f in ('submit_date', 'complete_date'):
                try:
                    v = v.strftime("%Y-%m-%d %H:%M")
                except Exception:
                    v = ''
            elif f == 'status':
                col = status_to_color.get(v)
                v = click.style(v, fg=col, bold=True)
            elif f == 'pod':
                v = job.details.get('pod_name', v)
            elif f == '#proc':
                v = sum(job.details.get('process_stats', {}).values())
            elif f == 'desc':
                v = job.description or ""
                if len(v) > 50:
                    v = v[:47] + "..."
            job_list.append(v)

        jobs_list.append(job_list)
    tbl = tabulate(jobs_list, headers=fields)
    if len(jobs_list) > 30 and is_page:
        click.echo_via_pager(tbl)
    else:
        click.echo(tbl)



import click
import hashlib
import json
import os
import sys
from click import echo
from tabulate import tabulate

from nextcodecli.utils import get_logger

log = get_logger('query')

PAGE_SIZE = 100000
MAX_ROWS_TO_DISPLAY = 100


def progress_callback(received, total):
    sys.stdout.write("  {:,} / {:,}    \r".format(received, total))


def get_results(qry, limit, filt, sort, output_file, is_json=False):
    row_count = qry.line_count or 0
    row_count_to_fetch = row_count
    if not output_file and not limit:
        limit = MAX_ROWS_TO_DISPLAY
        echo("No output file. Only fetching {} rows".format(limit))

    if limit:
        row_count_to_fetch = min(limit, row_count)

    if row_count != row_count_to_fetch:
        sys.stderr.write(
            "Retrieving top %s from %s rows of query results." % (row_count_to_fetch, row_count)
        )
    else:
        sys.stderr.write("Retrieving up to %s rows of query results." % (row_count_to_fetch))
    sys.stderr.flush()
    echo("")

    results = qry.get_results(
        row_count_to_fetch, 0, sort, filt, is_json=is_json, callback=progress_callback
    )

    if is_json:
        if output_file:
            with open(os.path.expanduser(output_file), 'w') as f:
                json.dump(results, f, indent=4)
            echo("Results have been written to '%s'" % output_file)
        else:
            echo(json.dumps(results, indent=4))
        return

    echo("")
    if output_file:
        with open(os.path.expanduser(output_file), 'w') as f:
            f.write(results)
        echo("Results have been written to '%s'" % output_file)

    else:
        rows = results.split("\n")
        rows = [r.split('\t') for r in rows]
        if len(rows) > MAX_ROWS_TO_DISPLAY:
            click.echo("Only showing top %s" % MAX_ROWS_TO_DISPLAY)
        tbl = tabulate(rows[1:MAX_ROWS_TO_DISPLAY], headers=rows[0])
        echo(tbl)



#!/usr/bin/env python
import click
import sys
import collections
from tabulate import tabulate
import dateutil
import yaml
from pathlib import Path

from nextcode.exceptions import ServerError
from nextcode.services.project.exceptions import ProjectError

from ...utils import print_tab, dumps, get_logger, abort

log = get_logger(name='commands.project', level='INFO')


@click.group(help="Project user management")
def users():
    pass


@users.command()
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
def list(ctx, is_raw):
    """
    List all users in this project
    """
    svc = ctx.obj.service
    try:
        users = svc.get_users_in_project()
    except ServerError as ex:
        abort(ex)
    if is_raw:
        print(dumps(users))
        return
    data = []
    for u in users:
        data.append((u["user_id"], u["user"], ", ".join(u["policies"])))
    tbl = tabulate(data, headers=["user_id", "user_name", "policies"])
    print(tbl)


@users.command()
@click.argument('user_name', required=True)
@click.option(
    '--roles', '-r', 'policies_string', default=None, help='Comma-separated list of policies'
)
@click.pass_context
def add(ctx, user_name, policies_string):
    """
    Add a user to the current project
    """
    svc = ctx.obj.service
    policies = []
    if policies_string:
        policies = [p.strip() for p in policies_string.split(",")]
    try:
        if policies:
            ret = svc.add_user_to_project(user_name=user_name, policies=policies)
        else:
            ret = svc.add_user_to_project(user_name=user_name)
    except ProjectError as ex:
        abort(ex)
    click.secho(
        f"User {user_name} has been added to project {ret['project']} with roles {', '.join(ret['policies'])}",
        bold=True,
    )


@users.command()
@click.argument('user_name', required=True)
@click.pass_context
def remove(ctx, user_name):
    """
    Remove a user from the current project
    """
    svc = ctx.obj.service
    try:
        svc.remove_user_from_project(user_name=user_name)
    except ProjectError as ex:
        abort(ex)
    click.secho(f"User {user_name} has been removed from project {svc.project_name}", bold=True)



#!/usr/bin/env python
import click
import os
from tabulate import tabulate

from nextcode.services.query.exceptions import MissingRelations, QueryError
from nextcode.exceptions import ServerError

from ...queryapi import get_results
from ...utils import get_logger, abort, print_tab, dumps
from . import check_project

log = get_logger(name='commands.query', level='INFO')


@click.command()
@click.option('-q', '--query', 'gor_query', help='Inline query to run')
@click.option('-j', '--jobtype', 'job_type', default="default", help='Job type to run the query')
@click.option('-f', '--queryfile', 'query_file', help='File which contains GOR query')
@click.option('-o', '--output', 'output_file', help='File to write results to')
@click.option('-l', '--limit', 'limit', default=0, help='Limit the number of rows returned')
@click.option(
    '-r', '--relations', 'relations', multiple=True, help='Virtual relation mapping for query'
)
@click.option(
    '-n',
    '--nowait',
    is_flag=True,
    default=False,
    help='Run command without waiting for it to finish',
)
@click.option(
    '-d', '--details', is_flag=True, default=False, help='Show progress details for running query'
)
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
@check_project
def run(
    ctx, gor_query, job_type, query_file, output_file, limit, relations, nowait, details, is_raw
):
    """
    Execute a GOR query via the Query API.

    You can either run a query directly or from a file.

    This command will automatically wait for the query
    to complete and fetch results but if you are running a long
    query you can use the --nowait option and then fetch results
    via the results command.

    """
    svc = ctx.obj.service

    prepared_relations = []
    for r in relations:
        lst = r.split("=", 1)
        name = lst[0]
        filename = lst[1]
        if not os.path.exists(filename):
            abort("File '{}' not found".format(filename))
        with open(filename) as f:
            data = f.read()
        prepared_relations.append({"name": name, "data": data})
    # options: (project, query)
    if query_file:
        with open(os.path.expanduser(query_file)) as f:
            gor_query = f.read()

    if not gor_query:
        click.secho("Specify --query or --queryfile", fg='red')
        return
    qry = None
    try:
        qry = svc.execute(gor_query, relations=prepared_relations, job_type=job_type, nowait=True)
        click.secho(
            "Running query {} against project {}...".format(qry.query_id, svc.project), bold=True
        )
        if not nowait:
            qry.wait()
    except MissingRelations as ex:
        click.secho(repr(ex), fg="red")
        txt = "You can add the relations with: "
        for r in ex.relations:
            txt += "-r {}=[filename]".format(r)
        click.echo(txt)
        return
    except KeyboardInterrupt:
        if qry:
            try:
                qry.cancel()
            except QueryError:
                pass
            abort("Query {} has been cancelled".format(qry.query_id))
        return None

    if qry.failed:
        click.secho(
            "Query {} failed with message: {}".format(qry.query_id, qry.status_message), fg="red"
        )
        return
    if is_raw:
        click.echo(dumps(qry.raw))
        return

    if qry.running:
        click.echo("Query {} is now running.".format(qry.query_id))
        click.echo("View query process with: nextcode query info {}".format(qry.query_id))
        return
    click.secho(
        "Query {} completed successfully in {} ms".format(qry.query_id, qry.duration), fg="green"
    )

    get_results(qry, limit, None, None, output_file)


@click.command()
@click.argument('project', nargs=1)
@click.pass_context
def project(ctx, project):
    """
    Set the current project.

    This command will set the project context in the client and save it
    into the current profile. A project is required to run any gor query
    or to inspect already run queries.
    """
    svc = ctx.obj.service
    svc.client.profile.project = project
    click.secho("Project has been set to: {}".format(project), bold=True)


@click.command()
@click.argument('query_id', nargs=1)
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
@check_project
def info(ctx, query_id, is_raw):
    """
    Retrieve information about a gor query by ID
    """
    svc = ctx.obj.service
    qry = svc.get_query(query_id)

    if is_raw:
        click.echo(dumps(qry.raw))
        return
    data = qry.raw
    print_tab("ID", data['id'])
    print_tab("Submitted", data['submit_date'])
    print_tab("User", data['user_name'])
    print_tab("Available", data['available'])
    print_tab("Template", data['template_name'] or "None (ad-hoc)")
    status = data['status']
    col = 'red'
    if status == 'DONE':
        col = 'green'
    elif status == 'RUNNING':
        col = 'yellow'
    print_tab("Status", click.style(status, fg=col, bold=True))

    if status == 'DONE':
        print_tab("Columns", data['stats']['column_count'])
        print_tab("Rows", data['stats']['line_count'])
        print_tab("Size", "%.1f MB" % (data['stats']['size_bytes'] / 1024 / 1024))
        run_time = data['stats']['end_timestamp'] - data['stats']['start_timestamp']
        queue_time = data['stats']['start_timestamp'] - data['stats']['submit_timestamp']
        print_tab("Queue Time", "%.2f sec" % (queue_time / 1000.0))
        print_tab("Run Time", "%.2f sec" % (run_time / 1000.0))
    elif status == 'FAILED':
        click.echo(data['error']['description'])


@click.command()
@click.argument('query_id', nargs=1)
@click.option(
    '-o',
    '--output',
    'output_file',
    help="Name of the output file for the results. Otherwise will dump a maximum of 100 rows to the console.",
)
@click.option(
    '-l', '--limit', 'limit', default=0, help="Maximum number of results to return (default all)"
)
@click.option(
    '-f',
    '--filter',
    'filt',
    help="GOR filter string to apply to the results. For example '-f \"chrom = 'chr1'\"",
)
@click.option(
    '-s',
    '--sort',
    'sort',
    help="Comma separated list of columns to sort on, including direction. Example: -s \"POS asc,Chrom desc\"",
)
@click.option('-j', '--json', 'is_json', is_flag=True, default=False)
@click.pass_context
@check_project
def results(ctx, query_id, output_file, limit, filt, sort, is_json):
    """
    Retrieve results from a completed GOR query.

    You can apply additional filters and sort options which
    are applied to the results before fetching.

    If output file is omitted only the first 100 lines are fetched.
    Use the --output option to fetch the entire result set.
    """
    svc = ctx.obj.service

    qry = svc.get_query(query_id)
    if not qry.done:
        abort("Query has status %s which is not DONE" % qry.status)

    get_results(qry, limit, filt, sort, output_file, is_json)


@click.command()
@click.argument('query_id', nargs=1)
@click.argument('filename', nargs=1)
@click.pass_context
@check_project
def download(ctx, query_id, filename):
    """
    Retrieve results from a completed GOR query to file.
    """
    svc = ctx.obj.service

    qry = svc.get_query(query_id)
    if not qry.done:
        abort("Query has status %s which is not DONE" % qry.status)

    line_count = qry.stats.get("line_count")
    click.secho(f"Downloading {line_count} lines to file...", bold=True)

    ret = qry.download_results(filename)
    click.secho(f"Query results have been downloaded to {ret}", bold=True)


@click.command()
@click.argument('query_id', nargs=1)
@click.pass_context
@check_project
def cancel(ctx, query_id):
    """
    Cancel a running GOR query
    """
    svc = ctx.obj.service

    qry = svc.get_query(query_id)
    try:
        qry.cancel()
    except QueryError as ex:
        abort(ex)
    click.secho("Query has been cancelled.", fg='yellow', bold=True)


@click.command()
@click.argument('query_id', nargs=1)
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@click.pass_context
@check_project
def progress(ctx, query_id, is_raw):
    """
    View the progress of a running query
    """
    svc = ctx.obj.service

    qry = svc.get_query(query_id)
    if qry.status != "RUNNING":
        abort("Query is not RUNNING")

    if is_raw:
        click.echo(dumps(qry.progress))
        return

    fields = ["job_id", "status", "fingerprint", "bytes", "lines"]

    rows = []
    for job_id, job in qry.progress.items():
        rows.append(
            [
                job_id,
                job['job-status'],
                job.get('fingerprint'),
                job.get('queryByteCount'),
                job.get('queryLineCount'),
            ]
        )
    tbl = tabulate(rows, headers=fields)
    click.echo(tbl)


@click.command()
@click.argument('query_id', nargs=1)
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@click.pass_context
@check_project
def details(ctx, query_id, is_raw):
    """
    View runtime details for a query
    """
    svc = ctx.obj.service

    qry = svc.get_query(query_id)
    url = qry.links['details']
    resp = svc.session.get(url)
    details = resp.json()['details']
    if not details:
        abort("No details found for query")

    if is_raw:
        click.echo(dumps(details))
        return
    error = details.get("error")
    if error is not None:
        abort("Error from server: %s" % error)

    request_id = details.get('requestId')
    if request_id is None:
        abort("Could not understand response from server: %s" % dumps(details))

    print_tab("requestId", request_id)
    click.secho("Tasks", bold=True)
    for task_name, task in details.get("tasks", {}).items():
        print(task_name)
        for k in sorted(task.keys()):
            print_tab(f" {k}", task[k])
        print("")


@click.command("list")
@click.option('-u', '--user', 'user_name', help='User to filter for')
@click.option('-s', '--status', 'status', default=None, help='Filter status')
@click.option(
    '-n', '--num', default=20, help='Maximum number of queries to return', show_default=True
)
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@click.option(
    '--all', 'is_all', is_flag=True, help='Fetch all queries in all projects (admin only)'
)
@click.pass_context
@check_project
def list_queries(ctx, user_name, status, num, is_raw, is_all):
    """
    List the queries that have been run by this user in the current project.

    To specify another user you can use the --user parameter but you must
    have the admin role on the server to do so. It is not possible to fetch
    queries for all users.
    """
    svc = ctx.obj.service
    if is_all and user_name:
        abort("You cannot use the --all flag and the --user parameter together")
    try:
        queries = svc.get_queries(limit=num, user_name=user_name, all=is_all)
    except ServerError as e:
        abort(e)
    if is_raw:
        click.echo(dumps([q.raw for q in queries]))
        return
    rows = []
    for qry in queries:
        rows.append(
            [
                qry.query_id,
                qry.submit_date.strftime("%Y-%m-%d %H:%M"),
                qry.status,
                qry.project_name,
                qry.user_name,
            ]
        )
    fields = ["query_id", "submitted", "status", "project", "user_name"]
    tbl = tabulate(rows, headers=fields)
    click.echo(tbl)


@click.command("wakeup")
@click.option(
    '-j',
    '--jobtype',
    'job_type',
    default="default",
    help='Job type to include with the wakeup request',
)
@click.option('-u', '--user', 'user_name', help='Optional user to include for tracking purposes')
@click.pass_context
@check_project
def wakeup(ctx, job_type, user_name):
    """
    Wake the server up.
    """
    svc = ctx.obj.service
    success = svc.wakeup(job_type, user_name)
    click.secho(f"Success: {success}", bold=True)



import os
import nextcode


def read_version():
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, 'VERSION'), 'r') as version_file:
        version = version_file.readline().strip()
        date = version_file.readline().strip()
        return version, date


__version__, version_date = read_version()


def get_version_string():
    format_str = "nextcode-cli/{version} ({version_date}) nextcode-sdk/{sdk_version}"
    version_msg = format_str.format(
        version=__version__, version_date=version_date, sdk_version=nextcode.__version__
    )
    return version_msg
